#ifndef TYPON_FUTURE_HPP_INCLUDED
#define TYPON_FUTURE_HPP_INCLUDED

#include <coroutine>
#include <memory>
#include <type_traits>
#include <utility>

#include <result.hpp>


namespace typon::rt
{

  template <typename Promise>
  using value_type = typename Promise::value_type;


  template <typename P>
  concept Void = std::is_void_v<value_type<P>>;


  template <typename P>
  concept Reference = std::is_reference_v<value_type<P>>;


  template <typename P>
  concept Complete = !Void<P> && !Reference<P>;


  template <typename P>
  concept Small = sizeof(value_type<P>) <= 2 * sizeof(void*);


  template <typename P>
  concept TriviallyCopyable = std::is_trivially_copyable_v<value_type<P>>;


  struct CoroutineGuard
  {
    std::coroutine_handle<> _coroutine;

    ~CoroutineGuard()
    {
      _coroutine.destroy();
    }
  };




  template <typename Promise>
  struct Future
  {
    template <typename T>
    struct always_false
    {
      static constexpr bool value { false };
    };

    static_assert(always_false<Promise>::value, "Unexpected Promise type for Future");
  };


  template <typename Promise>
  requires Complete<Promise> && (!Small<Promise>)
  struct Future<Promise>
  {
    std::coroutine_handle<Promise> _coroutine;
    bool _owning;

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      _coroutine = coroutine;
      _owning = ready;
      if (ready)
      {
        if (coroutine.promise()._exception)
        {
          std::rethrow_exception(coroutine.promise()._exception);
        }
      }
    }

    Future(const Future &) = delete;
    Future& operator=(const Future &) = delete;

    Future(Future && other) noexcept
      : _coroutine(other._coroutine)
      , _owning(std::exchange(other._owning, false))
    {}

    Future& operator=(Future && other) noexcept
    {
      std::swap(_owning, other._owning);
      std::swap(_coroutine, other._coroutine);
      return *this;
    }

    ~Future()
    {
      if (_owning)
      {
        _coroutine.destroy();
      }
    }

    auto get()
    {
      return _coroutine.promise().get();
    }
  };


  template <typename Promise>
  requires Complete<Promise> && Small<Promise> && (!TriviallyCopyable<Promise>)
  struct Future<Promise>
  {
    using value_type = Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;
    union
    {
      value_type _result;
    };

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        CoroutineGuard guard { coroutine };
        std::construct_at(&(_result), coroutine.promise().get());
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    Future(Future && other)
      noexcept(std::is_nothrow_move_constructible_v<value_type>)
    {
      _coroutine = other._coroutine;
      if (!_coroutine)
      {
        std::construct_at(std::addressof(_result), std::move(other._result));
      }
    }

    Future& operator=(Future && other)
      noexcept(std::is_nothrow_move_constructible_v<value_type>)
    {
      if (this != &other)
      {
        Future old { std::move(*this) };
        _coroutine = other._coroutine;
        if (!_coroutine)
        {
          std::construct_at(std::addressof(_result), std::move(other._result));
        }
      }
      return *this;
    }

    ~Future()
    {
      if (!_coroutine)
      {
        std::destroy_at(&(_result));
      }
    }

    auto get()
    {
      if (_coroutine)
      {
        return _coroutine.promise().get();
      }
      return _result;
    }
  };


  template <typename Promise>
  requires Complete<Promise> && Small<Promise> && TriviallyCopyable<Promise>
  struct Future<Promise>
  {
    using value_type = Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;
    union
    {
      value_type _result;
    };

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        CoroutineGuard guard { coroutine };
        std::construct_at(&(_result), coroutine.promise().get());
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    ~Future()
    {
      if (!_coroutine)
      {
        std::destroy_at(&(_result));
      }
    }

    auto get()
    {
      if (_coroutine)
      {
        return _coroutine.promise().get();
      }
      return _result;
    }
  };


  template <Reference Promise>
  struct Future<Promise>
  {
    using value_type = Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;
    std::remove_reference<value_type> * _result;

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        CoroutineGuard guard { coroutine };
        _result = std::addressof(coroutine.promise().get());
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    value_type get()
    {
      if (_coroutine)
      {
        return _coroutine.promise().get();
      }
      return *_result;
    }
  };


  template <Void Promise>
  struct Future<Promise>
  {
    using value_type = Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        CoroutineGuard guard { coroutine };
        coroutine.promise().get();
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    void get()
    {
      if (_coroutine)
      {
        _coroutine.promise().get();
      }
    }
  };

}


#endif // TYPON_FUTURE_HPP_INCLUDED
